
#ifndef CHANNEL_H
#define CHANNEL_H

#define FINISHED 1
#define SUCCESS 0
#define FAIL -1


/* This structure contains a number of function pointers describing
 * the data the channel will store.  Because we want the channel to
 * be general, we allow the channel client to specify the item type to be
 * stored when the channel is created.  The client does this by filling
 * in this structure and passing it to channel_create.
 *
 * The expected usage is as follows:
 *
 * - When the channel is created, an internal pool of items is created with item_create().
 * - When a write occurs, the client's item is copied into one of the items in the pool using 
 *   item_copy().
 * - When a read occurs, the item is copied from the pool into the client's item (the pointer 
 *    passed into read) using item_copy().
 * - When the channel is destroyed, the item_destroy() function should be called on
 *   all the items in the pool.
 *
 * Some caveats:
 * - item_create() should be considered very expensive (it will allocate memory), so it
 *   should be called as few times as possible.  Creating a pool at channel creation time
 *   is OK.  Creating an item every time channel_write() is called is not OK.
 *
 * - item_copy() should be considered expensive, and may block.  Locks should not be held
 *   when it is called (why?).
 */

struct channel_item_ops {
    void *(*item_create)(void);               /* Constructs an item.  Returns NULL on failure. */
    void (*item_copy)(void *dest, void *src); /* Copies an item to another item */
    void (*item_destroy)(void *item);         /* Cleans up any resources associated with the item */    
};

/**
 * channel_create: Create and initialise a channel.
 *
 * @nwriters: the number of threads which will write to the channel.
 * @nreaders: the number of threads which will read from the channel.
 * @size: the maximum number of items the channel may contain.
 * @item_ops: a struct of function pointers used to create, copy, and delete items.
 * @handle: a handle representing the channel.  Passed in to every function.
 *
 * This function creates and initialises a channel with @nwriters and
 * @nreaders, containing items defined by @ops.  The channel may store up
 * to @size writes; any writes after this will block until a read occurs.
 *
 * A handle to the channel is returned in @handle. 
 * 
 * Returns SUCCESS on success, FAIL on fail.
 */
int channel_create(int nreaders, int nwriters, int size,
		   struct channel_item_ops *ops,  /* Operations for the data items in this channel */
		   void **handle);                /* The handle to pass into the library for future 
						   * operations on this channel */
/**
 * channel_write: Write to the channel.
 *
 * @handle: the handle returned from channel_create.
 * @item: the item to write.
 *
 * This function copies @item and wakes up any waiting readers.
 * 
 * Returns SUCCESS on success, FAIL on fail.
 */
int channel_write(void *handle, void *item);


/**
 * channel_read: Read from the channel.
 *
 * @handle: the handle returned from channel_create.
 * @data: the buffer to read into.
 *
 * This function copies the next data item in the channel to @item.
 * 
 * Returns SUCCESS on success, FAIL on fail, FINISHED when no more
 * data is available (i.e. when all data has been read AND all writers
 * have finished)
 */
int channel_read(void *handle, void *item);

/**
 * channel_finished: Called by each writer when it has no more data to write.
 *
 * @handle: the handle returned from channel_create.
 *
 * This function informs the channel that the current thread has
 * finished writing.  If all writers have finished, the channel should
 * wake up any readers who are waiting.  When all readers have
 * finished the channel should be cleaned up.
 *
 */
void channel_finished(void *handle);

#endif /* CHANNEL_H */
